library oauth_twitter;

import 'dart:convert';
import 'dart:io';

import 'package:crypto/crypto.dart';
import 'package:oauth_twitter/src/oauth_db_sqlite.dart';
import 'package:url_launcher/url_launcher.dart';
import 'package:uuid/uuid.dart';

import 'src/twitter_api.dart';

/// A Calculator.
class OAuthTwitter {
  final String _apiKey;
  final String _apiSecret;
  final String _callBackUrl;
  final String _signatureMethod = "HMAC-SHA1";

  OAuthTwitter(this._apiKey, this._apiSecret, this._callBackUrl);

  /// Twitterにリクエストトークンを要求する
  Future<HttpClientResponse> oauthRequestToken() async {
    DateTime dtNow = DateTime.now();
    int unixTimeSec = (dtNow.millisecondsSinceEpoch / 1000).round();
    var uuid = Uuid().v4();

    var parameters = {
      "oauth_callback": _callBackUrl,
      "oauth_consumer_key": _apiKey,
      "oauth_signature_method": _signatureMethod,
      "oauth_timestamp": unixTimeSec.toString(),
      "oauth_nonce": uuid.toString(),
      "oauth_version": "1.0"
    };

    // うまく行く時も、常に空だったような気がする。。。
    // このリクエストによって取得するものなので、ここでは空にすることが必要？
    // キーごと空にしてもうまく行くんじゃないの...
    // TODO: 調査
    var secretByRequest = "";

    var signature = _createSignature(
        _apiSecret, secretByRequest, "POST", parameters, TwitterAPIs.URL_OAUTH_REQUEST_TOKEN);

    if (signature != null) {
      parameters["oauth_signature"] = signature;
    }

    var requestParams = _createRequestParams(parameters);
    // ex. "oauth_callback=yuma25689.ToriaezuTukutterZero%3A%2F%2F,oauth_consumer_key=xjpku6jywPx6uyfTczlpNXEyE,oauth_nonce=6134F0AE-0BF7-44E0-B76B-13BD807EE4D7,oauth_signature=u8KQMiIPn3aHGhGwdIjRjT4%2BUEs%3D,oauth_signature_method=HMAC-SHA1,oauth_timestamp=1628826747,oauth_version=1.0"

    var requestUrl = Uri.parse(TwitterAPIs.URL_OAUTH_REQUEST_TOKEN);

    var client = HttpClient();
    var ret = client.postUrl(requestUrl).then((HttpClientRequest request) {
      request.headers.add("Authorization", "OAuth " + requestParams);
      // Then call close.
      return request.close();
    });
    return ret;
  }

  /// RequestTokenで受信したレスポンスボディの内容を、一時データベースに保存
  void saveRequestTokenResponse(String responseBodyToken) {
    OAuthDataStorage storage = OAuthDataStorage();
    storage.deleteOAuthData(OAuthData.USERID_TEMP);
    final dicParams = urlParametersToDict(responseBodyToken);

    OAuthData data = OAuthData(OAuthData.USERID_TEMP);
    data.requestToken = dicParams["oauth_token"];
    data.requestTokenSecret = dicParams["oauth_token_secret"];
    storage.insertOAuthData(data);
  }

  /// OAuthの認証リクエストを行う(ブラウザで、Twitterが用意してくれる認証ページを開く
  /// それが終わると、Token取得時に指定したCALLBACK URLが呼び出されるので、
  /// そのCALLBACK URLでDeep Linkによりこのアプリが呼ばれるようにして、続きの処理を行う
  void oauthAuthentication(String responseBodyFromTokenRequest) {
    final dicParams = urlParametersToDict(responseBodyFromTokenRequest);
    String oauth_token = "";
    if (null != dicParams["oauth_token"]) {
      oauth_token = dicParams["oauth_token"]!;
    }

    var sURL = TwitterAPIs.URL_OAUTH_AUTHENTICATION + oauth_token;
    //var requestUrl = Uri.parse(TwitterAPIs.URL_OAUTH_AUTHENTICATION + oauth_token);

    _launchURL(sURL);
  }

  /// RequestTokenで受信したレスポンスボディの内容を、一時データベースに保存
  void saveAuthenticationResponse(String responseBodyAuthe) async {
    OAuthDataStorage storage = OAuthDataStorage();
    List<OAuthData> lstData = await storage.getOAuthData(OAuthData.USERID_TEMP);
    OAuthData data;
    if (0 < lstData.length) {
      data = lstData[0];
    } else {
      data = OAuthData(OAuthData.USERID_TEMP);
    }

    final dicParams = urlParametersToDict(responseBodyAuthe);

    data.authenticateToken = dicParams["oauth_token"];
    data.authenticateVerifier = dicParams["oauth_verifier"];
    data.authenticateDenied = dicParams["denied"];
    storage.updateOAuthData(data);
  }

  /// Twitterにアクセストークンを要求する
  /// 既に、tokenByAuthe(authentication)/verifierByAuthe/secretByRequestを取得済みでないとこの関数は通せない
  Future<HttpClientResponse> oauthAccessToken(
    String tokenByAuthe,
    String verifierByAuthe,
    String secretByRequest,
  ) async {
    DateTime dtNow = DateTime.now();
    int unixTimeSec = (dtNow.millisecondsSinceEpoch / 1000).round();
    var uuid = Uuid().v4();

    var parameters = {
      "oauth_consumer_key": _apiKey,
      "oauth_token": tokenByAuthe,
      "oauth_signature_method": _signatureMethod,
      "oauth_timestamp": unixTimeSec.toString(),
      "oauth_nonce": uuid.toString(),
      "oauth_verifier": verifierByAuthe,
      "oauth_version": "1.0"
    };

    var signature = _createSignature(
        _apiSecret, secretByRequest, "POST", parameters, TwitterAPIs.URL_OAUTH_ACCESS_TOKEN);

    if (signature != null) {
      parameters["oauth_signature"] = signature;
    }

    var requestParams = _createRequestParams(parameters);
    var requestUrl = Uri.parse(TwitterAPIs.URL_OAUTH_ACCESS_TOKEN);

    var client = HttpClient();
    var ret = client.postUrl(requestUrl).then((HttpClientRequest request) {
      request.headers.add("Authorization", "OAuth " + requestParams);
      // Then call close.
      return request.close();
    });
    return ret;
  }

  void _launchURL(url) async =>
      await canLaunch(url) ? await launch(url) : throw 'Could not launch $url';

  /// TwitterのOAuthに使用するSignatureを作成する
  String? _createSignature(String secretOfConsumer, String secretOfToken, String method,
      Map<String, String> params, String URL) {
    var ret = "";

    var secretOfConsumerEnc = encode2URL(secretOfConsumer);
    var secretOfTokenEnc = encode2URL(secretOfToken);
    var key = "$secretOfConsumerEnc&$secretOfTokenEnc";

    var arr = _encodeDictionaryParams(params);
    arr.sort((a, b) => a.compareTo(b));
    // query string作成
    var paramStringEnc = encode2URL(arr.join("&"));
    var methodEnc = encode2URL(method);
    var urlEnc = encode2URL(URL);
    var data = "$methodEnc&$urlEnc&$paramStringEnc";
    // ex. H6EZpRH4nkj85oeVimtrmNjnbMLAcWN7Jalk6NKVWRRVLmJ7Y5&
    var keyEnc = utf8.encode(key);
    var hmac_sha1 = Hmac(sha1, keyEnc);
    // ex. POST&https%3A%2F%2Fapi.twitter.com%2Foauth%2Frequest_token&oauth_callback%3Dyuma25689.ToriaezuTukutterZero%253A%252F%252F%26oauth_consumer_key%3Dxjpku6jywPx6uyfTczlpNXEyE%26oauth_nonce%3D123431FB-C302-4DE2-8392-E24F65143513%26oauth_signature_method%3DHMAC-SHA1%26oauth_timestamp%3D1628789559%26oauth_version%3D1.0
    var dataEnc = utf8.encode(data);
    var digest = hmac_sha1.convert(dataEnc);
    if (digest != null) {
      ret = base64.encode(digest.bytes);
      return ret;
    }
    return null;
  }

  /// AccessTokenで受信したレスポンスボディの内容を、UserIDをキーとしてデータベースに保存
  /// もしも、ResponseBodyにuser_idが含まれていなければ、false:エラーを返却するので注意する
  Future<bool> saveAccessTokenResponse(String responseBodyAccessToken) async {
    OAuthDataStorage storage = OAuthDataStorage();
    List<OAuthData> lstData = await storage.getOAuthData(OAuthData.USERID_TEMP);
    OAuthData data;
    if (0 < lstData.length) {
      data = lstData[0];
    } else {
      data = OAuthData(OAuthData.USERID_TEMP);
    }

    final dicParams = urlParametersToDict(responseBodyAccessToken);

    if (null == dicParams["user_id"]) {
      return false;
    }

    bool bUpdate = false;
    data.userId = dicParams["user_id"]!;
    List<OAuthData> lstDataUser = await storage.getOAuthData(data.userId);
    if (0 < lstDataUser.length) {
      data = lstDataUser[0];
      bUpdate = true;
    }
    data.accessToken = dicParams["oauth_token"];
    data.accessTokenSecret = dicParams["oauth_token_secret"];
    data.accessScreenName = dicParams["screen_name"];
    if (bUpdate) {
      storage.updateOAuthData(data);
    } else {
      storage.insertOAuthData(data);
    }

    return true;
  }

  /// Dictionary内のデータをURL形式にエンコードし、key=valueの形式の配列で返却
  List<String> _encodeDictionaryParams(Map<String, String> params) {
    List<String> parts = [];

    for (var entry in params.entries) {
      var keyString = encode2URL(entry.key);
      var valueString = encode2URL(entry.value);
      var query = "$keyString=$valueString";
      parts.add(query);
    }
    return parts;
  }

  /// Request用のパラメータ文字列を作成
  String _createRequestParams(Map<String, String> params) {
    var arr = _encodeDictionaryParams(params);
    arr.sort((a, b) => a.compareTo(b));
    var paramString = arr.join(",");
    return paramString;
  }

  /// RFC3986準拠で、URL形式にエンコードする
  String encode2URL(text) {
    return Uri.encodeQueryComponent(text);
  }

  /// keyA=valA&keyB=valB...の形式の文字列を辞書型に格納して返却する
  Map<String, String> urlParametersToDict(String params) {
    var dict = Map<String, String>();

    final tmp = params.split("&");
    for (var keyAndVal in tmp) {
      final tmp2 = keyAndVal.split("=");
      if (tmp2.length == 2) {
        dict[tmp2[0]] = tmp2[1];
      }
    }

    return dict;
  }

  Future<OAuthData?> getOneOAuthData(userId) async {
    OAuthDataStorage storage = OAuthDataStorage();
    List<OAuthData> lstData = await storage.getOAuthData(userId);
    OAuthData data;
    if (0 < lstData.length) {
      return lstData[0];
    }
    return null;
  }
}
