import 'package:path/path.dart';
import 'package:sqflite/sqflite.dart';

/// need dependencies
///  sqflite:
///  path:

/// 各ユーザー毎に、OAuth用のデータを保存するテーブル
/// ※ ただし、userId=-1を、認証中の一時領域として使用
class OAuthData {
  static const USERID_TEMP = "-1";

  /// ユーザーIDをキーとして保存するが、まだユーザーIDのわからない認証中は、一時バッファとして-1をユーザーIDとするので注意する
  String userId;
  String? requestToken;
  String? requestTokenSecret;

  String? authenticateToken;
  String? authenticateVerifier;
  String? authenticateDenied;

  String? accessToken;
  String? accessTokenSecret;
  String? accessScreenName;
  OAuthData(this.userId);

  /// データ内容をmapに変換して返却する
  Map<String, dynamic> toMap() {
    return {
      'userId': userId,
      'requestToken': requestToken,
      'requestTokenSecret': requestTokenSecret,
      'authenticateToken': authenticateToken,
      'authenticateVerifier': authenticateVerifier,
      'authenticateDenied': authenticateDenied,
      'accessToken': accessToken,
      'accessTokenSecret': accessTokenSecret,
      'accessScreenName': accessScreenName,
    };
  }

  /// mapからこのクラスにデータを変換して返却する
  static OAuthData fromMap(Map<String, dynamic> data) {
    var ret = OAuthData(data['userId']);
    ret.requestToken = data['requestToken'];
    ret.requestTokenSecret = data['requestTokenSecret'];
    ret.authenticateToken = data['authenticateToken'];
    ret.authenticateVerifier = data['authenticateVerifier'];
    ret.authenticateDenied = data['authenticateDenied'];
    ret.accessToken = data['accessToken'];
    ret.accessTokenSecret = data['accessTokenSecret'];
    ret.accessScreenName = data['accessScreenName'];
    return ret;
  }
}

class OAuthDataStorage {
  static const _DB_NAME = 'oauthInfo_database.db';

  /// OAuthのデータを保存するデータベースを使用するために開きその参照を返却する
  Future<Database> openStorage() async {
    return openDatabase(
      join(await getDatabasesPath(), _DB_NAME),
      onCreate: (db, version) {
        return db.execute(
          "CREATE TABLE OAuthData("
          " userId TEXT PRIMARY KEY,"
          " requestToken TEXT,"
          " requestTokenSecret TEXT,"
          " authenticateToken TEXT,"
          " authenticateVerifier TEXT,"
          " authenticateDenied TEXT,"
          " accessToken TEXT,"
          " accessTokenSecret TEXT,"
          " accessScreenName TEXT "
          " )",
        );
      },
      version: 1,
    );
  }

  /// 指定されたデータのOAuthDataをinsert
  Future<void> insertOAuthData(OAuthData data) async {
    final Database db = await openStorage();
    await db.insert(
      'OAuthData',
      data.toMap(),
      conflictAlgorithm: ConflictAlgorithm.replace,
    );
    db.close();
  }

  /// userIdで指定されたuserIdのOAuthDataを取得
  /// userIdが指定されていなければ、全件取得
  Future<List<OAuthData>> getOAuthData(String userId) async {
    final Database db = await openStorage();
    var wherePhrase = '';
    if (0 < userId.length) {
      wherePhrase = 'userId = ?';
    }
    final List<Map<String, dynamic>> maps =
        await db.query('OAuthData', where: wherePhrase, whereArgs: [userId]);
    db.close();
    return List.generate(maps.length, (i) {
      return OAuthData.fromMap(maps[i]);
    });
  }

  /// 指定されたデータのOAuthDataを更新
  Future<void> updateOAuthData(OAuthData data) async {
    final db = await openStorage();
    await db.update(
      'OAuthData',
      data.toMap(),
      where: "userId = ?",
      whereArgs: [data.userId],
      conflictAlgorithm: ConflictAlgorithm.fail,
    );
    db.close();
  }

  /// 指定されたuserIdのOAuthDataを削除
  Future<void> deleteOAuthData(String userId) async {
    final db = await openStorage();
    await db.delete('OAuthData', where: 'userId = ?', whereArgs: [userId]);
    db.close();
  }
}
